@@ -32,7 +32,6 @@
 #include <libsoup/soup-date.h>
 #include <uuid/uuid.h>
 #include <string.h>
-
 #include "gupnp-service.h"
 #include "gupnp-root-device.h"
 #include "gupnp-context-private.h"
@@ -59,9 +58,6 @@ struct _GUPnPServicePrivate {
         GQueue          *notify_queue;
 
         gboolean         notify_frozen;
-
-        GList           *pending_messages; /* Pending SoupMessages from this
-                                              service */
 };
 
 enum {
@@ -95,11 +91,37 @@ typedef struct {
         int           seq;
 
         guint         timeout_id;
+
+        GList        *pending_messages; /* Pending SoupMessages from this
+                                           subscription */
 } SubscriptionData;
 
 static void
 subscription_data_free (SubscriptionData *data)
 {
+        GUPnPContext *context;
+        SoupSession *session;
+
+        context = gupnp_service_info_get_context
+                        (GUPNP_SERVICE_INFO (data->service));
+        session = _gupnp_context_get_session (context);
+
+        /* Cancel pending messages */
+        while (data->pending_messages) {
+                SoupMessage *msg;
+
+                msg = data->pending_messages->data;
+
+                soup_session_cancel_message (session,
+                                             msg,
+                                             SOUP_STATUS_CANCELLED);
+
+                data->pending_messages =
+                        g_list_delete_link (data->pending_messages,
+                                            data->pending_messages);
+        }
+       
+        /* Further cleanup */
         while (data->callbacks) {
                 g_free (data->callbacks->data);
                 data->callbacks = g_list_delete_link (data->callbacks,
@@ -138,10 +160,24 @@ struct _GUPnPServiceAction {
         GString     *response_str;
 };
 
+GType
+gupnp_service_action_get_type (void)
+{
+        static GType our_type = 0;
+
+        if (our_type == 0)
+                our_type = g_pointer_type_register_static
+                                        ("GUPnPServiceAction");
+
+        return our_type;
+}
+
 /**
  * gupnp_service_action_get_name
  * @action: A #GUPnPServiceAction
  *
+ * Get the name of @action.
+ *
  * Return value: The name of @action
  **/
 const char *
@@ -156,22 +192,24 @@ gupnp_service_action_get_name (GUPnPServiceAction *action)
  * gupnp_service_action_get_locales
  * @action: A #GUPnPServiceAction
  *
- * Return value: An ordered (preferred first) #GList of locales preferred by
+ * Get an ordered (preferred first) #GList of locales preferred by
  * the client. Free list and elements after use.
+ *
+ * Return value: A #GList of #char* locale names.
  **/
 GList *
 gupnp_service_action_get_locales (GUPnPServiceAction *action)
 {
         g_return_val_if_fail (action != NULL, NULL);
 
-        return accept_language_get_locales (action->msg);
+        return http_request_get_accept_locales (action->msg);
 }
 
 /**
  * gupnp_service_action_get
  * @action: A #GUPnPServiceAction
  * @Varargs: tuples of argument name, argument type, and argument value
- * location, terminated with NULL.
+ * location, terminated with %NULL.
  *
  * Retrieves the specified action arguments.
  **/
@@ -246,26 +284,33 @@ gupnp_service_action_get_value (GUPnPServiceAction *action,
                                 GValue             *value)
 {
         xmlNode *node;
+        gboolean found;
 
         g_return_if_fail (action != NULL);
         g_return_if_fail (argument != NULL);
         g_return_if_fail (value != NULL);
 
+        found = FALSE;
         for (node = action->node->children; node; node = node->next) {
                 if (strcmp ((char *) node->name, argument) != 0)
                         continue;
 
-                gvalue_util_set_value_from_xml_node (value, node);
+                found = gvalue_util_set_value_from_xml_node (value, node);
 
                 break;
         }
+
+        if (!found)
+                g_warning ("Failed to retreive '%s' argument of '%s' action",
+                           argument,
+                           action->name);
 }
 
 /**
  * gupnp_service_action_set
  * @action: A #GUPnPServiceAction
  * @Varargs: tuples of return value name, return value type, and
- * actual return value, terminated with NULL.
+ * actual return value, terminated with %NULL.
  *
  * Sets the specified action return values.
  **/
@@ -376,7 +421,7 @@ gupnp_service_action_return (GUPnPServiceAction *action)
  * gupnp_service_action_return_error
  * @action: A #GUPnPServiceAction
  * @error_code: The error code
- * @error_description: The error description, or NULL if @error_code is
+ * @error_description: The error description, or %NULL if @error_code is
  * one of #GUPNP_CONTROL_ERROR_INVALID_ACTION,
  * #GUPNP_CONTROL_ERROR_INVALID_ARGS, #GUPNP_CONTROL_ERROR_OUT_OF_SYNC or
  * #GUPNP_CONTROL_ERROR_ACTION_FAILED, in which case a description is
@@ -557,8 +602,11 @@ query_state_variable (GUPnPService       *service,
 
 /* controlURL handler */
 static void
-control_server_handler (SoupServerContext *server_context,
-                        SoupMessage       *msg,
+control_server_handler (SoupServer        *soup_server,
+                        SoupMessage       *msg, 
+                        const char        *server_path,
+                        GHashTable        *query,
+                        SoupClientContext *soup_client,
                         gpointer           user_data)
 {
         GUPnPService *service;
@@ -567,12 +615,13 @@ control_server_handler (SoupServerContext *server_context,
         xmlDoc *doc;
         xmlNode *action_node;
         const char *soap_action, *action_name;
-        char *end, *date;
+        char *end;
+        gpointer response_body;
         GUPnPServiceAction *action;
 
         service = GUPNP_SERVICE (user_data);
 
-        if (strcmp (msg->method, SOUP_METHOD_POST) != 0) {
+        if (msg->method != SOUP_METHOD_POST) {
                 soup_message_set_status (msg, SOUP_STATUS_NOT_IMPLEMENTED);
 
                 return;
@@ -582,8 +631,8 @@ control_server_handler (SoupServerContext *server_context,
         client = GSSDP_CLIENT (context);
 
         /* Get action name */
-        soap_action = soup_message_get_header (msg->request_headers,
-                                               "SOAPAction");
+        soap_action = soup_message_headers_get (msg->request_headers,
+                                                "SOAPAction");
         action_name = strchr (soap_action, '#');
         if (!action_name) {
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
@@ -598,7 +647,8 @@ control_server_handler (SoupServerContext *server_context,
         *end = '\0';
 
         /* Parse action_node */
-        doc = xmlParseMemory (msg->request.body, msg->request.length);
+        doc = xmlParseMemory (msg->request_body->data,
+                              msg->request_body->length);
         action_node = xml_util_get_element ((xmlNode *) doc,
                                             "Envelope",
                                             "Body",
@@ -656,24 +706,21 @@ control_server_handler (SoupServerContext *server_context,
                          "</s:Body>"
                          "</s:Envelope>");
 
-        msg->response.owner  = SOUP_BUFFER_SYSTEM_OWNED;
-        msg->response.body   = g_string_free (action->response_str, FALSE);
-        msg->response.length = strlen (msg->response.body);
+        response_body = g_string_free (action->response_str, FALSE);
+
+        soup_message_set_response (msg,
+                                   "text/xml",
+                                   SOUP_MEMORY_TAKE,
+                                   response_body,
+                                   strlen (response_body));
 
         /* Cleanup */
         g_slice_free (GUPnPServiceAction, action);
 
         /* Server header on response */
-        soup_message_add_header (msg->response_headers,
-                                 "Server",
-                                 gssdp_client_get_server_id (client));
-
-        /* Date header on response */
-        date = soup_date_generate (time (NULL));
-        soup_message_add_header (msg->response_headers,
-                                 "Date",
-                                 date);
-        g_free (date);
+        soup_message_headers_append (msg->response_headers,
+                                     "Server",
+                                     gssdp_client_get_server_id (client));
 }
 
 /* Generates a standard (re)subscription response */
@@ -691,21 +738,14 @@ subscription_response (GUPnPService *service,
         client = GSSDP_CLIENT (context);
 
         /* Server header on response */
-        soup_message_add_header (msg->response_headers,
-                                 "Server",
-                                 gssdp_client_get_server_id (client));
-
-        /* Date header on response */
-        tmp = soup_date_generate (time (NULL));
-        soup_message_add_header (msg->response_headers,
-                                 "Date",
-                                 tmp);
-        g_free (tmp);
+        soup_message_headers_append (msg->response_headers,
+                                     "Server",
+                                     gssdp_client_get_server_id (client));
 
         /* SID header */
-        soup_message_add_header (msg->response_headers,
-                                 "SID",
-                                 sid);
+        soup_message_headers_append (msg->response_headers,
+                                     "SID",
+                                     sid);
 
         /* Timeout header */
         if (timeout > 0)
@@ -713,9 +753,9 @@ subscription_response (GUPnPService *service,
         else
                 tmp = g_strdup ("infinite");
 
-        soup_message_add_header (msg->response_headers,
-                                 "Timeout",
-                                 tmp);
+        soup_message_headers_append (msg->response_headers,
+                                     "Timeout",
+                                     tmp);
         g_free (tmp);
 
         /* 200 OK */
@@ -818,9 +858,9 @@ subscribe (GUPnPService *service,
         /* Add timeout */
         timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
-                data->timeout_id = g_timeout_add (timeout_seconds * 1000,
-                                                  subscription_timeout,
-                                                  data);
+                data->timeout_id = g_timeout_add_seconds (timeout_seconds,
+                                                          subscription_timeout,
+                                                          data);
         }
 
         /* Add to hash */
@@ -890,9 +930,9 @@ resubscribe (GUPnPService *service,
 
         timeout_seconds = parse_and_limit_timeout (timeout);
         if (timeout_seconds > 0) {
-                data->timeout_id = g_timeout_add (timeout_seconds * 1000,
-                                                  subscription_timeout,
-                                                  data);
+                data->timeout_id = g_timeout_add_seconds (timeout_seconds,
+                                                          subscription_timeout,
+                                                          data);
         }
 
         /* Respond */
@@ -913,8 +953,11 @@ unsubscribe (GUPnPService *service,
 
 /* eventSubscriptionURL handler */
 static void
-subscription_server_handler (SoupServerContext *server_context,
-                             SoupMessage       *msg,
+subscription_server_handler (SoupServer        *soup_server,
+                             SoupMessage       *msg, 
+                             const char        *server_path,
+                             GHashTable        *query,
+                             SoupClientContext *soup_client,
                              gpointer           user_data)
 {
         GUPnPService *service;
@@ -922,10 +965,10 @@ subscription_server_handler (SoupServerContext *server_context,
 
         service = GUPNP_SERVICE (user_data);
 
-        callback = soup_message_get_header (msg->request_headers, "Callback");
-        nt       = soup_message_get_header (msg->request_headers, "NT");
-        timeout  = soup_message_get_header (msg->request_headers, "Timeout");
-        sid      = soup_message_get_header (msg->request_headers, "SID");
+        callback = soup_message_headers_get (msg->request_headers, "Callback");
+        nt       = soup_message_headers_get (msg->request_headers, "NT");
+        timeout  = soup_message_headers_get (msg->request_headers, "Timeout");
+        sid      = soup_message_headers_get (msg->request_headers, "SID");
 
         /* Choose appropriate handler */
         if (strcmp (msg->method, GENA_METHOD_SUBSCRIBE) == 0) {
@@ -996,7 +1039,7 @@ gupnp_service_constructor (GType                  type,
         const GList *state_variables, *l;
         GUPnPContext *context;
         SoupServer *server;
-        SoupUri *uri;
+        SoupURI *uri;
         char *url;
 
         object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
@@ -1052,8 +1095,8 @@ gupnp_service_constructor (GType                  type,
         url = soup_uri_to_string (uri, TRUE);
         soup_uri_free (uri);
 
-        soup_server_add_handler (server, url, NULL,
-                                 control_server_handler, NULL, object);
+        soup_server_add_handler (server, url,
+                                 control_server_handler, object, NULL);
         g_free (url);
 
         /* Run listener on eventSubscriptionURL */
@@ -1064,8 +1107,8 @@ gupnp_service_constructor (GType                  type,
         url = soup_uri_to_string (uri, TRUE);
         soup_uri_free (uri);
 
-        soup_server_add_handler (server, url, NULL,
-                                 subscription_server_handler, NULL, object);
+        soup_server_add_handler (server, url,
+                                 subscription_server_handler, object, NULL);
         g_free (url);
 
         return object;
@@ -1152,8 +1195,6 @@ gupnp_service_dispose (GObject *object)
 {
         GUPnPService *service;
         GObjectClass *object_class;
-        GUPnPContext *context;
-        SoupSession *session;
 
         service = GUPNP_SERVICE (object);
 
@@ -1171,24 +1212,7 @@ gupnp_service_dispose (GObject *object)
         }
 
         /* Cancel pending messages */
-        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
-        if (context)
-                session = _gupnp_context_get_session (context);
-        else
-                session = NULL; /* Not the first time dispose is called. */
-
-        while (service->priv->pending_messages) {
-                SoupMessage *msg;
-
-                msg = service->priv->pending_messages->data;
-
-                soup_message_set_status (msg, SOUP_STATUS_CANCELLED);
-                soup_session_cancel_message (session, msg);
-
-                service->priv->pending_messages =
-                        g_list_delete_link (service->priv->pending_messages,
-                                            service->priv->pending_messages);
-        }
+        g_hash_table_remove_all (service->priv->subscriptions);
 
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
@@ -1281,7 +1305,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                               g_cclosure_marshal_VOID__POINTER,
                               G_TYPE_NONE,
                               1,
-                              G_TYPE_POINTER);
+                              GUPNP_TYPE_SERVICE_ACTION);
 
         /**
          * GUPnPService::query-variable
@@ -1304,7 +1328,8 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                               G_TYPE_NONE,
                               2,
                               G_TYPE_STRING,
-                              G_TYPE_POINTER);
+                              G_TYPE_POINTER /* Not G_TYPE_VALUE as this
+                                                is an outward argument! */);
 
         /**
          * GUPnPService::notify-failed
@@ -1333,7 +1358,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
  * gupnp_service_notify
  * @service: A #GUPnPService
  * @Varargs: Tuples of variable name, variable type, and variable value,
- * terminated with NULL.
+ * terminated with %NULL.
  *
  * Notifies listening clients that the properties listed in @Varargs
  * have changed to the specified values.
@@ -1355,9 +1380,9 @@ gupnp_service_notify (GUPnPService *service,
  * gupnp_service_notify_valist
  * @service: A #GUPnPService
  * @var_args: A va_list of tuples of variable name, variable type, and variable
- * value, terminated with NULL.
+ * value, terminated with %NULL.
  *
- * See gupnp_service_notify_valist(); this version takes a va_list for
+ * See gupnp_service_notify(); this version takes a va_list for
  * use by language bindings.
  **/
 void
@@ -1395,9 +1420,10 @@ gupnp_service_notify_valist (GUPnPService *service,
         }
 }
 
-/* Receiveid notify response. */
+/* Received notify response. */
 static void
-notify_got_response (SoupMessage *msg,
+notify_got_response (SoupSession *session,
+                     SoupMessage *msg,
                      gpointer     user_data)
 {
         SubscriptionData *data;
@@ -1409,16 +1435,21 @@ notify_got_response (SoupMessage *msg,
         data = user_data;
 
         /* Remove from pending messages list */
-        data->service->priv->pending_messages = 
-                g_list_remove (data->service->priv->pending_messages, msg);
+        data->pending_messages = g_list_remove (data->pending_messages, msg);
 
         if (SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
                 /* Success: reset callbacks pointer */
                 data->callbacks = g_list_first (data->callbacks);
+
+        } else if (msg->status_code == SOUP_STATUS_PRECONDITION_FAILED) {
+                /* Precondition failed: Cancel subscription */
+                g_hash_table_remove (data->service->priv->subscriptions,
+                                     data->sid);
+
         } else {
-                /* Notify failed */
+                /* Other failure: Try next callback or signal failure. */
                 if (data->callbacks->next) {
-                        SoupUri *uri;
+                        SoupURI *uri;
                         GUPnPContext *context;
                         SoupSession *session;
 
@@ -1430,10 +1461,8 @@ notify_got_response (SoupMessage *msg,
                         soup_uri_free (uri);
 
                         /* And re-queue */
-                        data->service->priv->pending_messages = 
-                                g_list_prepend
-                                        (data->service->priv->pending_messages,
-                                         msg);
+                        data->pending_messages = 
+                                g_list_prepend (data->pending_messages, msg);
 
                         context = gupnp_service_info_get_context
                                         (GUPNP_SERVICE_INFO (data->service));
@@ -1487,27 +1516,22 @@ notify_subscriber (gpointer key,
                 return;
         }
 
-        /* Add headers */
-        soup_message_add_header (msg->request_headers,
-                                 "Content-Type",
-                                 "text/xml");
+        soup_message_headers_append (msg->request_headers,
+                                     "NT",
+                                     "upnp:event");
 
-        soup_message_add_header (msg->request_headers,
-                                 "NT",
-                                 "upnp:event");
+        soup_message_headers_append (msg->request_headers,
+                                     "NTS",
+                                     "upnp:propchange");
 
-        soup_message_add_header (msg->request_headers,
-                                 "NTS",
-                                 "upnp:propchange");
-
-        soup_message_add_header (msg->request_headers,
-                                 "SID",
-                                 data->sid);
+        soup_message_headers_append (msg->request_headers,
+                                     "SID",
+                                     data->sid);
 
         tmp = g_strdup_printf ("%d", data->seq);
-        soup_message_add_header (msg->request_headers,
-                                 "SEQ",
-                                 tmp);
+        soup_message_headers_append (msg->request_headers,
+                                     "SEQ",
+                                     tmp);
         g_free (tmp);
 
         /* Handle overflow */
@@ -1517,13 +1541,14 @@ notify_subscriber (gpointer key,
                 data->seq = 1;
 
         /* Add body */
-        msg->request.owner  = SOUP_BUFFER_SYSTEM_OWNED;
-        msg->request.body   = g_strdup (property_set);
-        msg->request.length = strlen (property_set);
+        soup_message_set_request (msg,
+                                  "text/xml",
+                                  SOUP_MEMORY_TAKE,
+                                  g_strdup (property_set),
+                                  strlen (property_set));
 
         /* Queue */
-        data->service->priv->pending_messages = 
-                g_list_prepend (data->service->priv->pending_messages, msg);
+        data->pending_messages = g_list_prepend (data->pending_messages, msg);
 
         context = gupnp_service_info_get_context
                         (GUPNP_SERVICE_INFO (data->service));
@@ -1685,6 +1710,40 @@ strip_camel_case (char *camel_str)
         return stripped;
 }
 
+static GCallback
+find_callback_by_name (GModule    *module,
+                       const char *name)
+{
+        GCallback callback;
+        char *full_name;
+
+        /* First try with 'on_' prefix */
+        full_name = g_strjoin ("_",
+                               "on",
+                               name,
+                               NULL);
+
+        if (!g_module_symbol (module,
+                              full_name,
+                              (gpointer) &callback)) {
+                g_free (full_name);
+
+                /* Now try with '_cb' postfix */
+                full_name = g_strjoin ("_",
+                                       name,
+                                       "cb",
+                                       NULL);
+
+                g_module_symbol (module,
+                                 full_name,
+                                 (gpointer) &callback);
+        }
+
+        g_free (full_name);
+
+        return callback;
+}
+
 /* Use the strings from @name_list as details to @signal_name, and connect
  * callbacks with names based on these same strings to @signal_name::string. */
 static void
@@ -1703,47 +1762,28 @@ connect_names_to_signal_handlers (GUPnPService *service,
                 GCallback callback;
                 char     *callback_name;
                 char     *signal_detail;
-                char     *tmp;
 
                 signal_detail = (char *) name_node->data;
-                tmp = strip_camel_case (signal_detail);
+                callback_name = strip_camel_case (signal_detail);
 
                 if (callback_prefix) {
-                        callback_name = g_strjoin ("_",
-                                                   callback_prefix,
-                                                   tmp,
-                                                   NULL);
-                        g_free (tmp);
-                        tmp = callback_name;
-                }
+                        char *tmp;
+
+                        tmp = g_strjoin ("_",
+                                         callback_prefix,
+                                         callback_name,
+                                         NULL);
 
-                /* First try with 'on_' prefix */
-                callback_name = g_strjoin ("_",
-                                           "on",
-                                           tmp,
-                                           NULL);
-                if (!g_module_symbol (module,
-                                      callback_name,
-                                      (gpointer) &callback)) {
                         g_free (callback_name);
-                        /* Now try with '_cb' postfix */
-                        callback_name = g_strjoin ("_",
-                                                   tmp,
-                                                   "cb",
-                                                   NULL);
-                        if (!g_module_symbol (module,
-                                              callback_name,
-                                              (gpointer) &callback)) {
-                                g_free (tmp);
-                                g_free (callback_name);
-
-                                continue;
-                        }
+                        callback_name = tmp;
                 }
 
-                g_free (tmp);
+                callback = find_callback_by_name (module, callback_name);
                 g_free (callback_name);
 
+                if (callback == NULL)
+                        continue;
+
                 signal_detail = g_strjoin ("::",
                                            signal_name,
                                            signal_detail,
@@ -1762,23 +1802,24 @@ connect_names_to_signal_handlers (GUPnPService *service,
  * gupnp_service_signals_autoconnect
  * @service: A #GUPnPService
  * @user_data: the data to pass to each of the callbacks
- * @error: return location for a GError, or NULL
+ * @error: return location for a #GError, or %NULL
  *
- * A convenience function that attempts to connect all possible "action-invoked"
- * and "query-variable" signals to appropriate callbacks for the service
- * @service. It uses service introspection and GModule's introspective
- * features. It is very simillar to glade_xml_signal_autoconnect except that it
- * attempts to guess the names of the signal handlers on its own.
+ * A convenience function that attempts to connect all possible
+ * #GUPnPService::action-invoked and #GUPnPService::query-variable signals to
+ * appropriate callbacks for the service @service. It uses service introspection
+ * and GModule's introspective features. It is very simillar to
+ * glade_xml_signal_autoconnect() except that it attempts to guess the names of
+ * the signal handlers on its own.
  *
  * For this function to do its magic, the application must name the callback
- * functions for "action-invoked" signals by striping the CamelCase off the
- * action names and either prepend "on_" or append "_cb" to them. Same goes
- * for "query-variable" signals, except that "query_" should be prepended to
- * the variable name. For example, callback function for "GetSystemUpdateID"
- * action should be either named as "get_system_update_id_cb" or
- * "on_get_system_update_id" and callback function for the query of
- * "SystemUpdateID" state variable should be named "query_system_update_id_cb"
- * or "on_query_system_update_id".
+ * functions for #GUPnPService::action-invoked signals by striping the CamelCase
+ * off the action names and either prepend "on_" or append "_cb" to them. Same
+ * goes for #GUPnPService::query-variable signals, except that "query_" should
+ * be prepended to the variable name. For example, callback function for
+ * "GetSystemUpdateID" action should be either named as
+ * "get_system_update_id_cb" or "on_get_system_update_id" and callback function
+ * for the query of "SystemUpdateID" state variable should be named
+ * "query_system_update_id_cb" or "on_query_system_update_id".
  *
  * Note that this function will not work correctly if GModule is not supported
  * on the platform or introspection is not available for service @service.
